/*
 * Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 */

#ifndef __VIRTNET_DPA_COMMON_H__
#define __VIRTNET_DPA_COMMON_H__

#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <libutil/util.h>
#include <libutil/atomic.h>
#include <libflexio-dev/flexio_dev.h>
#include <virtnet_dpa_util.h>
#include <virtnet_dpa_common_dbg.h>
#include <virtnet_dpa_mru.h>

#define MAX_FETCHED_DESC	32
#define MAX_FETCHED_DESC_BIG_VQ	16
#define MAX_EHDR_LEN		142
#define MAX_VIRTQ_SIZE		256
#define VIRTNET_DPA_SYNC_TIMEOUT 1000
#define FLEXIO_QP_QPC_RQ_TYPE_ZERO_SIZE_RQ 0x3
#define BIT_MODE_BYTE_LOG 3
#define PAGE_SIZE_START_LOG 12
#define DESC_PER_BATCH		32
#define VIRTNET_DPA_DIM_CQ_COUNT 1024
#define MAX_MEM_DUMP_NUM_DWORD 1024
#define RQ_MIN_DESC_COUNT 16
#ifdef LATENCY_DEBUG
#define VIRTNET_DPA_EVENT_HANDLER_CTX_SIZE 1088
#else
#define VIRTNET_DPA_EVENT_HANDLER_CTX_SIZE 512
#endif

enum {
	/* hash result will be more evenly if table size is prime number */
	VIRTNET_DPA_AARFS_HASH_TABLE_SIZE = 65533,
	VIRTNET_DPA_AARFS_HASH_LOG_ENTRY_NUM_PER_DEV = 6,
	VIRTNET_DPA_AARFS_SYNC_SLEEP_US = 800000,
	VIRTNET_DPA_AARFS_DISABLE_SLEEP_S = 2,
	VIRTNET_DPA_AARFS_SYNC_ITERATIONS = 256,
	VIRTNET_DPA_AARFS_FQ_DEPTH = 256,
};

enum {
	VIRTNET_DPA_AARFS_FQ_STATE_EMPTY = 1,
	VIRTNET_DPA_AARFS_FQ_STATE_FULL = 2,
};

/* Copied from linux/virtio_net.h as dpa redefined virtio_net_hdr */
enum {
	VIRTIO_NET_F_CSUM = 0,
	VIRTIO_NET_F_GUEST_CSUM = 1,
	VIRTIO_NET_F_CTRL_GUEST_OFFLOADS = 2,
	VIRTIO_NET_F_MTU = 3,
	VIRTIO_NET_F_MAC = 5,
	VIRTIO_NET_F_GUEST_TSO4 = 7,
	VIRTIO_NET_F_GUEST_TSO6 = 8,
	VIRTIO_NET_F_GUEST_ECN = 9,
	VIRTIO_NET_F_GUEST_UFO = 10,
	VIRTIO_NET_F_HOST_TSO4 = 11,
	VIRTIO_NET_F_HOST_TSO6 = 12,
	VIRTIO_NET_F_HOST_ECN = 13,
	VIRTIO_NET_F_HOST_UFO = 14,
	VIRTIO_NET_F_MRG_RXBUF = 15,
	VIRTIO_NET_F_STATUS = 16,
	VIRTIO_NET_F_CTRL_VQ = 17,
	VIRTIO_NET_F_CTRL_RX = 18,
	VIRTIO_NET_F_CTRL_VLAN = 19,
	VIRTIO_NET_F_CTRL_RX_EXTRA = 20,
	VIRTIO_NET_F_GUEST_ANNOUNCE = 21,
	VIRTIO_NET_F_MQ = 22,
	VIRTIO_NET_F_CTRL_MAC_ADDR = 23,
	VIRTIO_NET_F_HOST_USO = 56,
	VIRTIO_NET_F_HASH_REPORT = 57,
	VIRTIO_NET_F_GUEST_HDRLEN = 59,
	VIRTIO_NET_F_RSS = 60,
	VIRTIO_NET_F_RSC_EXT = 61,
	VIRTIO_NET_F_STANDBY = 62,
	VIRTIO_NET_F_SPEED_DUPLEX = 63,
};

enum {
	VQ_DESC_LE_F_NEXT = (1 << 0),
	VQ_DESC_LE_F_WRITE = (1 << 1),
	VQ_DESC_LE_F_INDIRECT = (1 << 2),
};

enum {
	VQ_DESC_LE_F_AVAIL = (1 << 7),
	VQ_DESC_LE_F_USED = (1 << 15),
};

enum {
	/* MAX size is from MAX_DESC_CHAIN */
	VIRTNET_DPA_MAX_WQE_LOG_SIZE = 2,
	VIRTNET_DPA_RX_WQE_STRIDE_LOG_SIZE = 4,
	VIRTNET_DPA_MAX_RX_WQE_STRIDE_LOG_SIZE =
	(VIRTNET_DPA_RX_WQE_STRIDE_LOG_SIZE + VIRTNET_DPA_MAX_WQE_LOG_SIZE),
};

enum {
	LM_DIRTY_MAP_MODE_BIT     = 0x0,
	LM_DIRTY_MAP_MODE_BYTE    = 0x1,
};

enum {
	VIRTNET_DB_CQ_LOG_DEPTH = 2,
	VIRTNET_DPA_CVQ_SQ_DEPTH = 6,
	VIRTNET_DPA_CVQ_RQ_DEPTH = 6,
	/* One sqe_seg is 16B and One WQEBB is 64B.
	 * The algorithm to calculate sq size is as below:
	 * for each batch, the max sqe_seg number used is in LSO case,
	 * virtnet_dpa_ctrl_seg_set_lso() - 1 sqe_seg
	 * virtnet_dpa_eth_seg_set_lso() - 1 sqe_seg
	 * virtnet_dpa_inhdr_seg_set_lso() - 5 sqe_seg
	 * virtnet_dpa_data_segs_set_lso() - 4 sqe_seg
	 * A total of 11 DPA sqe_segs for LSO case which translates to
	 * up to 3 network WQEBBs and for a batch size of 32 we will need
	 * at least 66 network WQEBBs and rounding up
	 * it will be 128 network WQEBBs
	 */
	VIRTNET_NW_SQ_LOG_DEPTH = 7,
	VIRTNET_NW_SQ_CQ_LOG_DEPTH = 0,
	VIRTNET_FETCH_CQ_LOG_DEPTH = 3,
	VIRTNET_FETCH_SQ_LOG_DEPTH = 7,
	VIRTNET_DPA_TX_DESC_PACKED_LOG_DEPTH = 8,
	VIRTNET_DPA_TX_DESC_LOG_DEPTH = 8,
	VIRTNET_DPA_RX_DESC_LOG_DEPTH = 8,
	VIRTNET_DPA_RX_DESC_PACKED_LOG_DEPTH = 8,
};

enum {
	VIRTNET_MAX_TUNNEL_TX_POST_DESC_BUDGET = 1,
	VIRTNET_MAX_TX_POST_DESC_BUDGET = 256,
	VIRTNET_MAX_TX_INLINE_POST_DESC_BUDGET = 128,
	VIRTNET_MAX_RX_POST_DESC_BUDGET = 96,
	VIRTNET_MAX_RX_POLL_PKT_BUDGET = 96,
	VIRTNET_MAX_TX_POST_DESC_BUDGET_PACKED = 256,
	VIRTNET_MAX_RX_POST_DESC_BUDGET_PACKED = 256,
	VIRTNET_MAX_RX_POLL_PKT_BUDGET_PACKED = 256,
};

enum virtnet_queue_sizes {
	/* RM 3760006: will query total vq num after cap has it */
	VIRTNET_MAX_TOTAL_QUEUES = 8192,
	/* Hash table size should be prime number, it gives better distribution
	 * in the hash table statistically. 8209 is the closest prime number of
	 * VIRTNET_MAX_TOTAL_QUEUES(8k)
	 */
	TX_SCHED_HASH_TBL_SIZE = 8209,
};

#define VIRTNET_DPA_RX_AUX_HANDLER_POOL_SIZE 32
#define VIRTNET_DPA_TX_AUX_HANDLER_POOL_SIZE 32
/* The first 32 is vq-index based, remaining 32 is MRU(Most Recently
 * Used) based.
 */
#define VIRTNET_DPA_TX_DMA_QP_SIZE           64
/* dma_qp allocation with vq index, we call "fast" mode */
#define VIRTNET_DPA_DMA_QP_FAST_POOL_SIZE    32
#define VIRTNET_DPA_DMA_QP_MRU_POOL_SIZE     (VIRTNET_DPA_TX_DMA_QP_SIZE - \
					      VIRTNET_DPA_DMA_QP_FAST_POOL_SIZE)

#define VIRTNET_DPA_RX_DMA_QP_SIZE           64
/* dma_qp allocation with vq index, we call "fast" mode */
#define VIRTNET_DPA_RX_DMA_QP_FAST_POOL_SIZE    32
#define VIRTNET_DPA_RX_DMA_QP_MRU_POOL_SIZE     (VIRTNET_DPA_RX_DMA_QP_SIZE - \
					      VIRTNET_DPA_RX_DMA_QP_FAST_POOL_SIZE)

enum {
	VIRTIO_VERSION_NONE = 0,
	VIRTIO_VERSION_0_95 = 1,
	VIRTIO_VERSION_1_0 = 2,
};

enum {
	NET_HDR_LEN = 10,
	NET_HDR_MRG_RXBUF_LEN = 12,
	NET_HDR_HASH_REPORT_LEN = 18,
	NET_HDR_MRG_RXBUF_HASH_REPORT_LEN = 20,
};

enum dpa_sync_state_t{
	VIRTNET_DPA_SYNC_HOST_RDY = 1,
	VIRTNET_DPA_SYNC_DEV_RDY = 2,
};

enum {
	VIRTNET_DPA_AUX_OWNERSHIP = 0,
	VIRTNET_DPA_PRI_NETHDR_SKIP = 1,
	VIRTNET_DPA_PRI_NETHDR_POLL = 2,
};

enum virtnet_outbox_type {
	EMU_MNG_OUTBOX = 0,
	SF_OUTBOX = 1,
};

enum virtnet_vq_mode {
	SPLIT = 0,
	PACKED = 1,
};

enum handler_types {
	VIRTNET_INVALID_HANDLER = 0,
	VIRTNET_SQ_DB_HANDLER = 1,
	VIRTNET_SQ_AUX_HANDLER = 2,
	VIRTNET_RQ_DB_HANDLER = 3,
	VIRTNET_RQ_AUX_HANDLER = 4,
	VIRTNET_NW_RQ_HANDLER = 5,
	VIRTNET_CTRL_VQ_HANDLER = 6,
	VIRTNET_ADMIN_VQ_HANDLER = 7,
	VIRTNET_TX_SCHED_POLL_HANDLER = 8,
	VIRTNET_TX_SCHED_EVENT_HANDLER = 9,
};

enum {
	VIRTNET_DPA_STR_SZ = 32,
};

struct refcount {
	struct spinlock_s lock;
	atomic32_t cnt;
	uint8_t valid;		/* when not set, try_get will abort */
	char last_caller[VIRTNET_DPA_STR_SZ];
	uint8_t err;
};

struct virtnet_dpa_cq {
	uint32_t cq_num;
	uint32_t log_cq_depth;
	flexio_uintptr_t cq_ring_daddr;
	flexio_uintptr_t cq_dbr_daddr;
	struct flexio_cq *cq;
};

struct virtnet_dpa_error_handler_attr {
	flexio_uintptr_t event_handler_ctx;
	enum handler_types event_handler_type;
};

struct virtnet_dpa_vq_counters {
	uint64_t received_desc;
	uint64_t completed_desc;
	/* dma_q_used is tx/rx specific */
	uint32_t dma_q_used;
	uint32_t handler_schd_num;
	uint16_t max_post_desc_num;
	uint8_t batch_num;
	uint32_t aux_handler_schd_num;
	uint32_t nw_handler_schd_num;
	uint64_t total_bytes;
	uint32_t err_handler_schd_num;
};

struct virtnet_dpa_vq_pkt_cnt {
	/* packets of size <= 64 bytes */
	uint64_t bytes_0_64;
	/* packets of size 65 bytes to 127 bytes */
	uint64_t bytes_65_127;
	/* packets of size 128 bytes to 255 bytes */
	uint64_t bytes_128_255;
	/* packets of size 256 bytes to 511 bytes */
	uint64_t bytes_256_511;
	/* packets of size 512 bytes to 1023 bytes */
	uint64_t bytes_512_1023;
	/* packets of size 1024 bytes to 1522 bytes */
	uint64_t bytes_1024_1522;
	/* packets of size 1523 bytes to 2047 bytes */
	uint64_t bytes_1523_2047;
	/* packets of size 2048 bytes to 4095 bytes */
	uint64_t bytes_2048_4095;
	/* packets of size 4096 bytes to 8191 bytes */
	uint64_t bytes_4096_8191;
	/* packets of size 8192 bytes to 9022 bytes */
	uint64_t bytes_8192_9022;
};

struct virtnet_dpa_rx_shadow_avail_ring {
	uint16_t avail[POW2(VIRTNET_DPA_RX_DESC_LOG_DEPTH)];
};

struct virtnet_dpa_shadow_avail_ring {
	uint16_t avail[POW2(VIRTNET_DPA_TX_DESC_LOG_DEPTH)];
};

struct virtnet_dpa_vq_desc {
	uint64_t addr;
	uint32_t len;
	uint16_t flags;
	uint16_t next;
};

struct virtnet_dpa_packed_vq_desc{
	uint64_t addr;
	uint32_t len;
	uint16_t id;
	uint16_t flags;
};

/* For packed VQ, id, len and flags are in the last 8 bytes. To enable
 * optimized completion of desc defining struct in qwords.
 */
struct virtnet_dpa_vq_desc_qword {
	uint64_t qword0;
	uint64_t qword1;
};

union virtnet_dpa_packed_vq_shadow_desc {
	uint64_t data;
	struct {
		uint32_t len;
		uint16_t id;
		uint16_t flags;
	};
};

struct virtnet_dpa_cq_ctx {
	uint32_t cqn;
	uint32_t ci;
	struct virtnet_dev_cqe64 *ring;
	struct virtnet_dev_cqe64 *cqe;
	uint32_t comp_wqe_idx;
	uint32_t *dbr;
	uint8_t hw_owner_bit;
};

struct virtnet_dpa_ring_ctx {
	uint32_t num;
	/* Stores the q number which is right shifted by 8 bits to directly
	 * write into the WQE
	 */
	uint32_t num_shift;
	union flexio_dev_sqe_seg *ring;
	uint32_t wqe_seg_idx;
	uint32_t *dbr;
	uint32_t pi;
	uint32_t ci;
};

/* virtnet_dpa_vq_state values:
 *
 * @VIRTNET_DPA_VQ_STATE_INIT - VQ is created, but cannot handle doorbells.
 * @VIRTNET_DPA_VQ_STATE_SUSPEND - VQ is suspended, no outgoing DMA, can be restarted.
 * @VIRTNET_DPA_VQ_STATE_RDY - Can handle doorbells.
 * @VIRTNET_DPA_VQ_STATE_ERR - VQ is in error state.
 */
enum virtnet_dpa_vq_state {
	VIRTNET_DPA_VQ_STATE_INIT = 1 << 0,
	VIRTNET_DPA_VQ_STATE_RDY = 1 << 1,
	VIRTNET_DPA_VQ_STATE_SUSPEND = 1 << 2,
	VIRTNET_DPA_VQ_STATE_ERR = 1 << 3,
};

enum virtq_types {
	VIRTNET_DPA_INVALID = 0,
	VIRTNET_DPA_SX = 1,
	VIRTNET_DPA_RX = 2,
	VIRTNET_DPA_CTRL_Q = 3,
	VIRTNET_DPA_ADMIN_Q = 4,
};

/* For mergeable buffer feature, post 2 desc per wqe for non jumbo
 * and 8 per wqe for Jumbo mode.
 */
enum {
	MRG_BUF_NON_JUMBO_WQE_COUNT = 2,
	MRG_BUF_JUMBO_WQE_COUNT = 8,
};

enum {
	MRG_BUF_NON_JUMBO_WQE_STRIDE = 32,
	MRG_BUF_JUMBO_WQE_STRIDE = 128,
};

struct virtnet_dpa_avail_ring {
	uint64_t base_addr;
	uint16_t next_index;
	uint16_t done;
	uint16_t tun_posted_index;
	uint16_t reserved;
};

struct virtnet_dpa_used_ring {
	uint64_t base_addr;
	uint16_t next_index;
	uint16_t reserved[3];
};

struct virtnet_dpa_desc_table {
	uint64_t base_addr;
};

struct virtnet_dpa_packed_desc_ring {
	uint64_t base_addr;
	uint16_t num_desc;
	uint16_t done;
	uint32_t reserved;
};

struct virtnet_dpa_splitq {
	struct virtnet_dpa_avail_ring avail_ring;
	struct virtnet_dpa_desc_table desc_table;
	struct virtnet_dpa_used_ring used_ring;
};

struct virtnet_dpa_packedq {
	struct virtnet_dpa_packed_desc_ring desc_ring;
};

struct virtnet_window_dev_config {
	uint32_t mkey;
	flexio_uintptr_t haddr;
	flexio_uintptr_t heap_memory;
} __attribute__((__packed__, aligned(8)));

struct virtnet_dpa_mem_dump_attr {
	uint32_t mkey;
	flexio_uintptr_t haddr;
	uint32_t emu_crossing_mkey;
	uint64_t addr;
	int num_dword;
} __attribute__((__packed__, aligned(8)));

struct virtnet_window_hw_idx_attr {
	uint16_t hw_available_index;
	flexio_uintptr_t heap_memory;
} __attribute__((__packed__, aligned(8)));

struct virtnet_dpa_aarfs_dump_attr {
	flexio_uintptr_t ctx_addr;
	uint32_t mkey;
	flexio_uintptr_t host_entries;
	flexio_uintptr_t host_mru_head;
	flexio_uintptr_t host_stats;
} __attribute__((__packed__, aligned(8)));

struct virtnet_dpa_vq_ctx {
	struct virtnet_dpa_splitq splitq;
	struct virtnet_dpa_packedq packedq;
	uint32_t sf_crossing_mkey;
	uint32_t emu_crossing_mkey;
	uint32_t dumem_mkey;
	bool need_wait_sq_cq;
} __attribute__((__packed__, aligned(8)));

struct virtnet_dpa_sqrq {
	struct virtnet_dpa_vq_ctx vq_ctx;
	uint8_t wqe_log_size;
};

struct virtnet_dpa_common_heap {
	/* Rq packet len/byte statistics memory base address */
	flexio_uintptr_t pkt_cnt_addr;
	/* aarfs ctx memory base address */
	flexio_uintptr_t aarfs_ctx_addr;
};

/* This struct must be same as virtnet_prov_ctrl_offloads */
struct virtnet_dpa_offload_features {
	uint8_t rx_csum;
};

struct virtnet_dpa_vq_features {
	struct virtnet_dpa_offload_features offloads;
	uint8_t tx_csum;
	uint8_t tso_ipv4;
	uint8_t tso_ipv6;
	uint8_t hdr_len;
	uint8_t notify_data;
	uint8_t hash_report;
	uint8_t mrg_rxbuf;
	uint8_t ring_packed;
};

struct virtnet_dpa_dev_info {
	uint16_t is_pf:1;
	uint16_t rsvd:3;
	uint16_t dev_id:12;
};

struct virtnet_dpa_device_ctx {
	struct virtnet_dpa_common_heap *heap;
	uint16_t *used_idx_cache;
	struct virtnet_dpa_vq_features features;
	struct virtnet_dpa_dev_info dev_info;
	bool pkt_cnt_enable;
	bool aarfs_enable;
};

#define CYCLE_NAME_LEN 16
struct virtnet_dpa_cpu_counter {
	uint32_t cycles;
	uint32_t events;
	uint64_t total_cycles;
};

#define MAX_CPU_COUNTERS 8
struct virtnet_dpa_cpu_counters {
	struct virtnet_dpa_cpu_counter counter[MAX_CPU_COUNTERS];
};

struct virtio_hdr_flags_gso_type {
	uint8_t flags;
	uint8_t gso_type;
} __attribute__((packed, aligned(2)));

union virtio_net_hdr_u16 {
	uint16_t flags_gso_type_raw;
	struct virtio_hdr_flags_gso_type type;
};

/* Smaller virtio_net_hdr for tx side which doesn't need to use fields starting
 * csum_start. We only need first 6 bytes, adding 2 bytes for padding so we can
 * copy hdr in one instruction.
 */
struct virtio_net_hdr_tx {
	union virtio_net_hdr_u16 flags_gso;
	uint16_t hdr_len;
	uint16_t gso_size;
	uint16_t rsvd;
} __attribute__((packed, aligned(2)));

struct virtnet_dpa_vnet_hdr_rctx {
	struct virtio_net_hdr_tx hdrs[VIRTNET_MAX_TX_POST_DESC_BUDGET];
};

union virtnet_len_key {
	uint64_t data;
	struct {
		be32_t len;
		be32_t key;
	};
};

struct virtnet_tx_dma_q {
	be32_t vnet_hdr_rctx_mkey;
	struct virtnet_dpa_vnet_hdr_rctx *vnet_hdr_rctx;
	struct virtnet_dpa_shadow_avail_ring *avail_ring;
};

struct virtnet_rx_dma_q {
	struct virtnet_dpa_rx_shadow_avail_ring *avail_ring;
};

/* This will contain VIRTNET_DPA_TX_DMA_QP_SIZE +
 * VIRTNET_DPA_RX_DMA_QP_SIZE
 */
struct virtnet_dpa_dma_q_state {
	uint32_t qp_num;
	uint8_t qp_in_error;
};

struct virtnet_dma_q {
	struct flexio_qp *qp;
	struct virtnet_dpa_cq_ctx fetch_cq_ctx;
	struct virtnet_dpa_vq_desc *desc_table;
	struct virtnet_dpa_packed_vq_desc *packed_desc_ring;

	flexio_uintptr_t qp_dbr_daddr;
	flexio_uintptr_t qp_sq_daddr;

	uint32_t hw_qp_sq_pi;
	uint32_t qp_num;
	uint16_t hw_qp_depth;
	uint8_t err_status;
	atomic32_t fast_pool_used;
	be32_t virtnet_shadow_vq_mkey;
	be32_t packed_vq_shadow_mkey;
	union {
		struct virtnet_tx_dma_q tx_q;
		struct virtnet_rx_dma_q rx_q;
	};
	uint64_t health_addr;
	uint32_t health_lkey;
};

struct virtnet_dpa_dirtymap_para {
	uint32_t vhost_log_page:5;	/* Log (base 2) of page size */
	uint32_t mode:2;		/* 0-bitmap. 1-bytemap */
	uint32_t pad:25;		/* padding */
	uint32_t mkey;			/* mkey to log (a.k.a dirty_bitmap) */
	uint64_t addr;			/* addr of dirty_bitmap */
};

struct virtnet_dpa_shared_mem_ctx {
	uint16_t depth[VIRTNET_MAX_TX_POST_DESC_BUDGET];
	uint16_t avail[VIRTNET_MAX_TX_POST_DESC_BUDGET];
	/* Number of desc to process per batch */
	uint8_t batch_flag[VIRTNET_MAX_TX_POST_DESC_BUDGET/DESC_PER_BATCH];
};

struct virtnet_dpa_rq_shadow_elem {
	/* All rq specific shadow memory should be put here */
	uint16_t used_idx;
};

struct virtnet_rq_wqe_shadow_ring_jumbo {
	/* need only first desc to update net_hdr */
	uint64_t desc_addr;
	uint16_t desc_len[MRG_BUF_JUMBO_WQE_COUNT];
	uint16_t desc_idx[MRG_BUF_JUMBO_WQE_COUNT];
};

struct virtnet_rq_wqe_shadow_ring_non_jumbo {
	/* need only first desc to update net_hdr */
	uint64_t desc_addr;
	uint16_t desc_len[MRG_BUF_NON_JUMBO_WQE_COUNT];
	uint16_t desc_idx[MRG_BUF_NON_JUMBO_WQE_COUNT];
};

struct virtnet_rq_wqe_shadow_ring_elem {
	/* need only first desc to update net_hdr */
	uint64_t desc_addr;
	struct  {
		uint16_t desc_len;
		uint16_t desc_idx;
	} idx_len[];
};

struct virtnet_rq_wqe_shadow_ring {
	flexio_uintptr_t base_addr;
};

struct virtnet_dpa_packed_ctx {
	bool wrap_counter;
	uint16_t used_flags;
	uint16_t ctrl_desc_num;
	union virtnet_dpa_packed_vq_shadow_desc *shadow_ring;
	uint32_t msix_vector;
};

struct virtnet_dpa_event_handler_ctx {
	struct virtnet_dpa_device_ctx *dev_ctx;
	uint8_t net_hdr_len;
	enum virtq_types type;
	uint8_t batch;
	uint8_t is_chain;
	struct refcount refcount;
	union virtnet_len_key len_key;

	struct virtnet_dpa_cq_ctx guest_db_cq_ctx;
	struct virtnet_dpa_cq_ctx nw_cq_ctx;
	struct virtnet_dpa_ring_ctx ring_ctx;
	struct virtnet_dpa_vq_counters eh_vq_counters;
	struct virtnet_dpa_vq_counters *host_vq_counters;

	struct virtnet_dpa_sqrq sqrq;
	uint32_t emu_mng_gvmi;
	uint32_t sf_gvmi;
	uint32_t emu_db_to_cq_id;
	uint32_t msix_cqn;
	uint16_t db_eu_running;
	flexio_uintptr_t window_base_addr;
	uint16_t vq_index;
	uint16_t vq_depth;
	enum virtnet_vq_mode mode;
	struct virtnet_dpa_packed_ctx packed_ctx;
#ifdef LATENCY_DEBUG
	struct virtnet_dpa_cpu_counters cpu_cnts;
	struct virtnet_dpa_latency_stats lat_stats;
#endif
	struct {
		/* Primary handler sets these fields before invoking auxiliary
		 * handler.
		 */
		uint8_t aux_handler_running;
		uint8_t pri_handler_running;
		uint8_t aux_handler_bailout;
		uint8_t avail_req_poll_done;
		uint8_t desc_req_poll_done;
	} aux_handler;
	struct {
		struct virtnet_dpa_aux_handler_ctx *aux_ctx;
		struct virtnet_dma_q *dma_q;
	} error_handling;
	/* This will tell us if we have net_hdr data updated */
	bool use_qp_based_net_hdr;
	uint8_t cur_outbox_type;
	uint8_t stats_clear;
	/* Unique global RQ identifier */
	uint16_t rq_id;
	struct virtnet_dpa_dirtymap_para dirty_map_para;
	struct virtnet_dpa_shared_mem_ctx *shared_mem_ctx;
	uint64_t dim_stats_addr;
	uint32_t dim_mr_lkey;
	bool dirty_log_enable;
	uint32_t cqe_err;
	uint8_t last_syndrome;
	uint8_t last_vendor_err_synd;
	uint8_t emu_ctx_pi_valid;
	bool first_sched;
	/* It's an array with size of qsize */
	struct virtnet_dpa_rq_shadow_elem *rq_shadow;
	struct virtnet_rq_wqe_shadow_ring rq_wqe_shadow;
};

struct virtnet_dpa_tunnel_event_handler_ctx {
	struct virtnet_dpa_event_handler_ctx ctx;
	struct {
		struct virtnet_dpa_cq qp_rqcq;
		uint32_t hw_qp_sq_pi;
		uint32_t hw_qp_cq_ci;
		uint32_t hw_qp_depth;
		uint16_t max_tunnel_desc;
		uint32_t qp_num;
		flexio_uintptr_t qp_sq_buff;
		flexio_uintptr_t qp_rq_buff;
		flexio_uintptr_t dbr_daddr;
	} dma_qp;
};

struct virtnet_dpa_vq_data {
	uint32_t dump_mem[MAX_MEM_DUMP_NUM_DWORD];
	struct virtnet_dpa_vq_counters vq_counters;
	struct virtnet_dpa_vq_pkt_cnt pkt_cnt;
	struct virtnet_dpa_event_handler_ctx event_ctx;
	enum dpa_sync_state_t state;
	flexio_uintptr_t db_cq_ring_addr;
	uint8_t err;
} __attribute__((__packed__, aligned(8)));

struct virtnet_dpa_vq_hw_idx_data {
	uint16_t hw_available_index;
} __attribute__((__packed__, aligned(8)));

/* DIM stats per ms */
struct virtnet_dpa_vq_dim_ms {
	uint64_t ppms;
	uint64_t bpms;
	uint64_t epms;
};

struct virtnet_dpa_vq_dim_stats_sample {
	uint64_t time;
	uint64_t num_pkts;
	uint64_t num_bytes;
	uint64_t num_events;
};

struct virtnet_dpa_vq_dim_stats {
	struct flexio_cq *cq;

	uint8_t tune_state;
	uint8_t steps_left;
	uint8_t steps_right;
	uint8_t tired;
	uint8_t profile_ix;
	volatile uint8_t in_use;
	uint32_t count;


	struct virtnet_dpa_vq_dim_ms old_stats_per_ms;
	struct virtnet_dpa_vq_dim_stats_sample old_stats;
	struct virtnet_dpa_vq_dim_stats_sample new_stats;
};

struct virtnet_dpa_msix_send {
	uint32_t sf_gvmi;
	uint32_t cqn;
	uint32_t emu_mng_gvmi;
};

struct virtnet_dpa_aux_handler_ctx {
	/* Primary handler will update the following that
	 * auxiliary handler will read to fetch desc
	 */
	struct virtnet_dpa_event_handler_ctx *pri_ctx;
	struct virtnet_dma_q *dma_q;
	/* Primary handler will use this cqn to wake up auxiliary handler */
	uint32_t cqn;
	uint32_t eu_num;
	uint32_t avail_ring_start_idx;
	uint32_t desc_count;
	/* When a secondary handler is allocated, by the caller,
	 * in_use is set, indicating it is allocated. When usage is done,
	 * its cleared (freed) for next usage.
	 */
	volatile bool in_use;
};

typedef void (virtnet_func_t) (void);
struct virtnet_dpa_rx_aux_handler {
	struct flexio_event_handler *db_handler;
	struct virtnet_dpa_cq db_cq;
	/* Points to virtnet_dpa_rx_aux_handler_ctx structure. */
	flexio_uintptr_t rx_aux_ctx;
};

struct virtnet_dpa_rx_aux_handler_attr {
	flexio_uintptr_t rx_aux_ctx_pool;
	flexio_uintptr_t rx_aux_stack_addr;
	uint32_t rx_aux_pool_size;
};

struct virtnet_dpa_tx_aux_handler {
	struct flexio_event_handler *db_handler;
	struct virtnet_dpa_cq db_cq;
	/* Points to virtnet_dpa_aux_handler_ctx structure. */
	flexio_uintptr_t tx_aux_ctx;
};

struct virtnet_dpa_tx_aux_handler_attr {
	flexio_uintptr_t tx_aux_ctx_pool;
	flexio_uintptr_t tx_aux_stack_addr;
	uint32_t tx_aux_pool_size;
};

struct virtnet_dpa_tx_dma_q_dev_access {
	struct virtnet_dma_q qps[VIRTNET_DPA_TX_DMA_QP_SIZE];
	flexio_uintptr_t stack_daddr;
};

struct virtnet_dpa_rx_dma_q_dev_access {
	struct virtnet_dma_q qps[VIRTNET_DPA_RX_DMA_QP_SIZE];
	flexio_uintptr_t stack_daddr;
};

struct virtnet_dev_cqe64 {
	uint32_t rsvd0[3];	/* 00h..06h - Reserved */
	uint32_t rx_hash_result;/* 03h - RX hash function result */
	uint8_t rx_hash_type;	/* 04h 24..31 - RX hashing performed */
	uint8_t rsvd17[3];	/* 04h 0..23 - Reserved */
	uint32_t rsvd20[2];	/* 05h..06h - Reserved */
	uint8_t csum_ok;	/* 07h 24..26 - checksum ok bits */
	uint8_t l4_hdr_type;	/* 07h 20..22 - l4 hdr type bits */
	uint8_t rsvd30[2];	/* 07h 0..15 - Reserved */
	uint32_t srqn_uidx;	/* 08h - SRQ number or user index */
	uint32_t rsvd36[2];	/* 09h..0Ah - Reserved */
	uint32_t byte_cnt;	/* 0Bh - Byte count */
	uint32_t rsvd48[2];	/* 0Ch..0Dh - Reserved */
	uint32_t qpn;		/* 0Eh - QPN */
	uint16_t wqe_counter;	/* 0Fh 16..31 - WQE counter */
	uint8_t signature;	/* 0Fh 8..15 - Signature */
	uint8_t op_own;		/* 0Fh 0 - Ownership bit */
} __attribute__((packed, aligned(8)));

struct virtnet_dpa_aarfs_session {
	uint16_t src_port;		/* swap them on rx path */
	uint16_t dst_port;		/* swap them on rx path */
	uint16_t dev_id;
};

struct virtnet_dpa_aarfs_hash_entry {
	MRU_ENTRY(uint16_t) mru_entry;
	struct virtnet_dpa_aarfs_session session;
	uint16_t q_idx;			/* q index used to steer rx packet */
	uint16_t total_pkt_cnt;		/* update on every lso */
	uint16_t shadow_pkt_cnt;	/* update on timer check */
};

struct virtnet_dpa_aarfs_stats {
	uint32_t hash_insert_failed;
	uint32_t hash_conflict;
	uint32_t dev_flows_full;
	uint32_t add_queue_full;
	uint32_t del_queue_full;
};

struct virtnet_dpa_aarfs_ctx {
	struct spinlock_s lock;		/* lock to protect add/removal */
	struct virtnet_dpa_aarfs_hash_entry *entries;
	struct virtnet_dpa_aarfs_stats stats;
	atomic32_t *entries_counter;	/* per dev counter of entries */
	MRU_HEAD(, uint16_t) mru_head;
	uint64_t add_queues_addr;
	uint64_t del_queue_addr;
	uint32_t q_mkey;
};

struct virtnet_dpa_aarfs_fq_entry {
	uint16_t src_port;		/* swap them on rx path */
	uint16_t dst_port;		/* swap them on rx path */
	uint16_t dev_id;
	uint16_t q_idx;			/* q index used to steer rx packet */
};

/* flow queue, used to add or delete flow steering rule */
struct virtnet_dpa_aarfs_fq {
	struct virtnet_dpa_aarfs_fq_entry entries[VIRTNET_DPA_AARFS_FQ_DEPTH];
	uint32_t pi;
	uint32_t ci;
};

static inline int
virtnet_dpa_event_ctx_entry_size(void)
{
	int size = sizeof(struct virtnet_dpa_event_handler_ctx);
	_Static_assert(sizeof(struct virtnet_dpa_event_handler_ctx)
		       <= VIRTNET_DPA_EVENT_HANDLER_CTX_SIZE,
		       "sizeof(struct virtnet_dpa_event_handler_ctx) exceeds "
		       "expected max length");
	/* Use sizeof(struct virtnet_dpa_event_ctx_entry) and align to 64B cache
	 * line and derive the size.
	 */
	return 64 * DIV_ROUND_UP(size, 64);
}

static inline uint32_t
virtnet_dpa_aarfs_port_hash(uint16_t src_port, uint16_t dst_port, uint16_t dev_id)
{
	uint32_t hash;
	src_port = src_port | (src_port >> 8);
	dst_port = dst_port | (dst_port >> 8);
	hash = dst_port | src_port;
	hash = hash & POW2MASK(VIRTNET_DPA_AARFS_HASH_LOG_ENTRY_NUM_PER_DEV);

	/* cluster them in nearby cache line for the same dev */
	hash = hash | (dev_id << VIRTNET_DPA_AARFS_HASH_LOG_ENTRY_NUM_PER_DEV);
	hash = hash % VIRTNET_DPA_AARFS_HASH_TABLE_SIZE;
	return hash;
}
#endif
