/*
 * Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 */

#ifndef __VIRTNET_DPA_TX_SCHED_H__
#define __VIRTNET_DPA_TX_SCHED_H__

#include <libflexio/flexio.h>

extern flexio_func_t virtnet_dpa_tx_sched_rpc_hdlr;
extern flexio_func_t virtnet_dpa_tx_sched_hdlr;
extern flexio_func_t virtnet_dpa_tx_sched_poll_rpc_hdlr;
extern flexio_func_t virtnet_dpa_tx_sched_poll_hdlr;
extern flexio_func_t virtnet_dpa_tx_sched_event_rpc_hdlr;
extern flexio_func_t virtnet_dpa_tx_sched_event_hdlr;
extern flexio_func_t virtnet_dpa_tx_sched_dev2host_copy;

enum virtnet_dpa_tx_sched_type {
	VIRTNET_DPA_TX_SCHED,
	VIRTNET_DPA_TX_SCHED_POLL,
	VIRTNET_DPA_TX_SCHED_EVENT,
	VIRTNET_DPA_TX_SCHED_TYPE_MAX
};

/* Common handler structure for tx poll/event/scheduler handler */
struct virtnet_dpa_tx_sched_hdlr_obj {
	struct flexio_event_handler *tx_sched_hdlr;
	struct virtnet_dpa_cq trigger_hdlr_cq;
};

struct virtnet_dpa_tx_sched_hdlr_attr {
	enum virtnet_dpa_tx_sched_type type;
	struct virtnet_dpa_tx_sched_hdlr_obj *obj;
	flexio_func_t *hdlr_func;
	union {
		flexio_uintptr_t daddr;
		struct virtnet_dpa_tx_sched_hdlr_ctx *sched;
		struct virtnet_dpa_tx_sched_poll_hdlr_ctx *poll;
		struct virtnet_dpa_tx_sched_event_hdlr_ctx *event;
	} hdlr_ctx;
	union {
		struct virtnet_dpa_tx_sched_poll_work *poll;
		struct virtnet_dpa_tx_sched_event_wq *event;
	} work_entry;
	flexio_uintptr_t err_attr;
};

struct virtnet_dpa_tx_sched_hash_entry {
	flexio_uintptr_t daddr;
	uint32_t emu_db_to_cq_id;
	uint16_t used;
};

struct virtnet_dpa_tx_sched_hash_info {
	struct virtnet_dpa_tx_sched_hash_entry entry[TX_SCHED_HASH_TBL_SIZE];
	pthread_mutex_t lock;
};

struct virtnet_dpa_tx_sched_poll_ctx {
	struct virtnet_dpa_tx_sched_hdlr_obj obj[TX_SCHED_POLL_HDLR_POOL_MAX_SZ];
	/* All poll handlers ctx allocate at once. Points to first poll
	 * handler ctx
	 */
	struct virtnet_dpa_tx_sched_poll_hdlr_ctx *hdlr_ctx;
	struct virtnet_dpa_tx_sched_poll_rpc_ctx *rpc_ctx;
	flexio_uintptr_t err_attr;

	uint8_t hdlr_core_idx;
	uint8_t hdlr_num;
};

struct virtnet_dpa_tx_sched_event_ctx {
	struct virtnet_dpa_tx_sched_hdlr_obj obj[TX_SCHED_EVENT_HDLR_POOL_SZ];
	/* All device event handlers ctx allocate at one time. This points to
	 * the first device event handler ctx
	 */
	struct virtnet_dpa_tx_sched_event_hdlr_ctx *hdlr_ctx;
	struct virtnet_dpa_tx_sched_event_rpc_ctx *rpc_ctx;
	flexio_uintptr_t err_attr;

	uint8_t hdlr_eu_pos;
	uint8_t hdlr_core_idx;
};

struct virtnet_dpa_tx_sched_ctx {
	struct virtnet_dpa_tx_sched_hash_info hash_info;

	struct virtnet_dpa_tx_sched_hdlr_obj obj;
	struct virtnet_dpa_tx_sched_hdlr_ctx *hdlr_ctx;
	struct virtnet_dpa_tx_sched_rpc_ctx *rpc_ctx;
	flexio_uintptr_t db_cq_ring_buffer;
	struct virtnet_dpa_tx_sched_data *sched_data;
	struct ibv_mr *sched_data_mr;
	uint8_t hdlr_core_idx;
	/* When enabled tx scheduler, if core num of sched EU[5] is more
	 * than the total of scheduler handler and poll handler number. The
	 * rest of cores will assign to rx. e.g. If dpa_core_start is 0,
	 * dpa_core_end is 10 and poll handler number is 5. core[0]EU[5] is
	 * assigned to scheduler handler. core[1-5]EU[5] is assigned to
	 * poll handler. core[6-10]EU[5] is assigned to rx. This ensures rx
	 * and tx get similar number of EUs to use
	 */
	bool share_eu;

	struct virtnet_dpa_tx_sched_poll_ctx poll_ctx;
	struct virtnet_dpa_tx_sched_event_ctx event_ctx;
};

int
virtnet_dpa_tx_sched_init(struct virtnet_dpa_tx_sched_ctx **tx_sched_ctx_p,
			  struct flexio_process *process,
			  struct ibv_context *emu_mgr_ibv_ctx,
			  struct virtnet_dpa_ctx *dpa_ctx,
			  struct ibv_pd *emu_mgr_ibv_pd,
			  const struct virtnet_prov_init_attr *attr);
void
virtnet_dpa_tx_sched_uninit(struct virtnet_dpa_tx_sched_ctx *tx_sched_ctx,
			    struct flexio_process *process,
			    bool tx_sched_enabled);
void
virtnet_dpa_tx_sched_stats_query(void *handler, cJSON *stats);
#endif
